"""
Author: Lukas Runt
Email: lrunt@students.zcu.cz
Date: November 06, 2023
Version: 1.0.0
Description: Knowledge base
"""

import BazeDat


class Graph:
    """
    Represent the structure of graph

    Attributes
    ----------
    nodes: List[Node]
        list of nodes in graph
    """

    def __init__(self):
        self.nodes = []

    def add_node(self, node):
        """
        Method for adding node
        :param node: Node what will be added to the graph
        """
        self.nodes.append(node)

    def add_edge(self, node1: int, node2: int):
        """
        Method adds edge between two nodes
        :param node1: Number of node which will be connected with node2
        :param node2: Number of node which will be connected with node1
        """
        self.nodes[node1].neighbours.append(node2)
        self.nodes[node2].neighbours.append(node1)

    def has_edge(self, node1: int, node2: int):
        """
        Method controls if there is an edge between two nodes
        :param node1: Number of node 1
        :param node2: Number of node 2
        :return: True - Edge exists, False - Edge do not exist
        """
        if node2 in self.nodes[node1].neighbours and node1 in self.nodes[node2].neighbours:
            return True
        else:
            return False


class Node:
    """
    Represents one vertex of the graph

    Attributes
    ----------
    neighbours: List[int]
        list of neighbours represented by index of the node in the graph
    value: int
        numeric value of the node
    defined: bool
        True - value is defined by user, False - value is not defined
    grid: List[int]
        indexes of fields which are in same grid 3x3
    """

    def __init__(self):
        self.neighbours = []
        self.value = 0
        self.defined = False
        self.grid = []


class Sudoku:
    """
    Class representing sudoku

    Attributes
    ----------
    graph: Graph
        is undirected graph of sudoku
    number_of_filled_fields: int
        number of fields which still wasn't filled
    possible_colors: List[List[int]]
        is list with possible colors for every node in the graph
    stack: List[List[]]
        is list where the progress is saved
    """

    def __init__(self):
        self.graph = Graph()
        self.number_of_filled_fields = 0
        self.possible_colors = []
        self.stack = []
        self.graph_init()

    def graph_init(self):
        """
        Creating graph of sudoku play board
        """
        for i in range(0, BazeDat.NUMBER_OF_TILES):
            self.graph.add_node(Node())

        # Connections in row
        for i in range(BazeDat.HEIGHT_OF_PLAY_BOARD):
            for j in range(BazeDat.WIDTH_OF_PLAY_BOARD):
                for k in range(j + 1, BazeDat.WIDTH_OF_PLAY_BOARD):
                    self.graph.add_edge(i * BazeDat.WIDTH_OF_PLAY_BOARD + j, i * BazeDat.WIDTH_OF_PLAY_BOARD + k)

        # Connections in column
        for i in range(BazeDat.WIDTH_OF_PLAY_BOARD):
            for j in range(BazeDat.HEIGHT_OF_PLAY_BOARD):
                for k in range(j + 1, BazeDat.HEIGHT_OF_PLAY_BOARD):
                    self.graph.add_edge(i + j * BazeDat.HEIGHT_OF_PLAY_BOARD, i + k * BazeDat.HEIGHT_OF_PLAY_BOARD)

        # Connections in grid
        for grid in BazeDat.grids:
            for i in range(len(grid)):
                for j in range(i + 1, len(grid)):
                    if not self.graph.has_edge(grid[i], grid[j]):
                        self.graph.add_edge(grid[i], grid[j])

        self.add_grid_to_nodes()

    def add_grid_to_nodes(self):
        """
        Adding other fields in grid to node
        """
        for grid in BazeDat.grids:
            for i in grid:
                self.graph.nodes[i].grid = grid

    def fill_graph_with_given_information(self):
        """
        Method filling the sudoku with data
        """
        self.number_of_filled_fields = 0
        for i in range(BazeDat.NUMBER_OF_TILES):
            self.graph.nodes[i].value = BazeDat.sudoku[i]
            if BazeDat.sudoku[i] != 0:
                self.number_of_filled_fields += 1
                self.graph.nodes[i].defined = True
            else:
                self.graph.nodes[i].defined = False

    def control_input_data(self):
        """
        Controlling if given data are valid,
        if there is same color in neighbour, then returns false
        :return: True - is valid/no conflict between colors, False - No valid data/there is a conflict
        """
        for i in range(len(self.graph.nodes)):
            value_of_node = self.graph.nodes[i].value
            if value_of_node != 0:
                for neighbour in self.graph.nodes[i].neighbours:
                    if value_of_node == self.graph.nodes[neighbour].value:
                        return False
        return True

    def find_solution_of_sudoku(self):
        """
        Solving of sudoku
        :return: true - sudoku have a solution, False - sudoku do not have solution
        """
        self.find_all_possible_colors_for_all_fields()
        while self.number_of_filled_fields != BazeDat.NUMBER_OF_TILES:
            min_length_index = min(
                (idx for idx, arr in enumerate(self.possible_colors) if arr and len(arr) > 0),
                key=lambda x: len(self.possible_colors[x])
            )
            self.add_value_to_node(min_length_index, self.possible_colors[min_length_index][0],
                                   self.possible_colors[min_length_index])
            self.number_of_filled_fields += 1
            self.update_possible_colors(min_length_index)
            # Backtracking if there is no solution after assigning value to field
            if not self.control_solution_exists():
                if not self.backtracking():
                    return False
        return True

    def backtracking(self):
        """
        Backtracking of algorithm, we have colored nodes bad
        We are going back and trying other colors
        :return: True - alternative choice exists, False - we have tried all chooses -> result do not exist
        """
        if len(self.stack) == 0:
            return False
        else:
            previous_step = self.stack.pop()
            node = previous_step[0]
            value = previous_step[1]
            self.reverse_possible_colors(node)
            self.number_of_filled_fields -= 1
            index_of_value = self.possible_colors[node].index(value)
            if index_of_value < len(self.possible_colors[node]) - 1:
                self.add_value_to_node(node, self.possible_colors[node][index_of_value + 1], self.possible_colors[node])
                self.number_of_filled_fields += 1
                self.update_possible_colors(node)
                return True
            else:
                if self.backtracking():
                    return True
                else:
                    return False

    def add_value_to_node(self, node, value, possible_colors_for_field):
        """
        Adding value to the node of the graph and adding this ter to the stack
        :param possible_colors_for_field: possible colors for the field counted with sudoku rules
        :param node: Node which have new value
        :param value: color of the node
        """
        self.graph.nodes[node].value = value
        self.stack.append([node, value, possible_colors_for_field])

    def control_solution_exists(self):
        """
        Checking if there is solution for sudoku
        :return: True - solution probably exists, False - solution do not exist
        """
        number_of_empty_arrays = sum(1 for arr in self.possible_colors if len(arr) == 0)
        if number_of_empty_arrays == self.number_of_filled_fields:
            return True
        else:
            return False

    def reverse_possible_colors(self, node):
        """
        Method to reverting the step
        :param node: node what we are reverting
        """
        self.graph.nodes[node].value = 0
        self.possible_colors[node] = self.find_possible_colors(node)
        for neighbour in self.graph.nodes[node].neighbours:
            if self.graph.nodes[neighbour].value == 0:
                self.possible_colors[neighbour] = self.find_possible_colors(neighbour)

    def update_possible_colors(self, node):
        """
        Method updating possible colors after adding value to one node
        :param node: node to which the color were assigned
        """
        self.possible_colors[node] = []
        for neighbour in self.graph.nodes[node].neighbours:
            if self.graph.nodes[node].value in self.possible_colors[neighbour]:
                self.possible_colors[neighbour].remove(self.graph.nodes[node].value)

    def find_all_possible_colors_for_all_fields(self):
        """
        Method finds all possible colors for all play fields of sudoku
        """
        self.possible_colors = []
        for i in range(len(self.graph.nodes)):
            if self.graph.nodes[i].value != 0:
                self.possible_colors.append([])
            else:
                self.possible_colors.append(self.find_possible_colors(i))

    def find_possible_colors(self, node):
        """
        Method computes all possible colors for node
        :param node: node to which we will compute the possible colors
        :return: what colors the node can have without breaking rules of sudoku
        """
        possible_numbers = list(BazeDat.all_colors)
        for neighbour in self.graph.nodes[node].neighbours:
            value = self.graph.nodes[neighbour].value
            if value in possible_numbers:
                possible_numbers.remove(value)
        return possible_numbers

    def reset_sudoku(self):
        """
        Method resets solution of sudoku
        """
        self.stack = []
        self.possible_colors = []
        self.number_of_filled_fields = 0
        for node in self.graph.nodes:
            node.value = 0
            node.defined = False

    def toStr(self):
        """
        Text representation of sudoku
        """
        text_representation = ""
        for i in range(BazeDat.HEIGHT_OF_PLAY_BOARD):
            for j in range(BazeDat.WIDTH_OF_PLAY_BOARD):
                value = self.graph.nodes[i * BazeDat.WIDTH_OF_PLAY_BOARD + j].value
                text_representation += f"{value} "

                # Add vertical separator after every 3 numbers in a row
                if (j + 1) % 3 == 0 and j < BazeDat.WIDTH_OF_PLAY_BOARD - 1:
                    text_representation += "| "

            text_representation += "\n"

            # Add horizontal separator after every 3 rows
            if (i + 1) % 3 == 0 and i < BazeDat.HEIGHT_OF_PLAY_BOARD - 1:
                text_representation += "-" * (BazeDat.WIDTH_OF_PLAY_BOARD * 2 + 6)
                text_representation += "\n"
        return text_representation
