"""
Author: Lukas Runt
Email: lrunt@students.zcu.cz
Date: November 06, 2023
Version: 1.0.0
Description: Inference module, managing communication between modules
"""
import sys

from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget
from qtpy import QtGui

import BazeDat
import BazeZnalosti
import KomunikacniModul
import VysvetlovaciModul


def app():
    """
    Method creating main window of the application
    """
    application = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(application.exec_())


class MainWindow(QMainWindow):
    """
    Starting main window of application
    """

    def __init__(self):
        super().__init__()

        self.sudoku = BazeZnalosti.Sudoku()

        self.setWindowTitle(KomunikacniModul.main_window_title)
        self.setWindowIcon(QtGui.QIcon(KomunikacniModul.main_window_icon))
        self.setGeometry(100, 100, 1000, 500)

        self.central_widget = QWidget(self)
        self.scene = KomunikacniModul.SudokuGUI()

        self.setCentralWidget(self.central_widget)
        self.central_widget.setLayout(self.scene.layout)
        self.scene.solve_button.clicked.connect(self.solution_button_action)
        self.scene.explanation_button.clicked.connect(self.explanation_button_action)

    def solution_button_action(self):
        """
        Choose which action will be performed
        """
        if self.scene.solve_button.text() == KomunikacniModul.solve_button_text:
            if self.solve_sudoku():
                self.scene.change_solve_button_text()
        else:
            self.reset_play_board()
            self.scene.change_solve_button_text()

    def explanation_button_action(self):
        """
        Handling the explanation button action
        """
        if self.scene.list_widget.currentText() == KomunikacniModul.print_process:
            explanation_text = VysvetlovaciModul.explain_solution_procedure(self.sudoku.stack)
            self.scene.explanation_area.setText(explanation_text)
        else:
            coordination = self.scene.coordination_list.currentText()
            explanation_text = VysvetlovaciModul.explain_tile_value_choice(coordination, self.sudoku)
            self.scene.explanation_area.setText(explanation_text)

    def solve_sudoku(self):
        """
        Sending data from graphical interface to database
        :return: True - Sudoku solved, False - Solving sudoku failed
        """
        for i in range(len(self.scene.text_boxes)):
            value = self.scene.text_boxes[i].text()
            if value == "" or value == " ":
                BazeDat.sudoku[i] = 0
            else:
                BazeDat.sudoku[i] = int(value)
        self.sudoku.fill_graph_with_given_information()
        if not self.sudoku.control_input_data():
            KomunikacniModul.show_error_message(KomunikacniModul.error_bad_input)
            return False
        if not self.sudoku.find_solution_of_sudoku():
            KomunikacniModul.show_error_message(KomunikacniModul.error_no_solution)
            return False
        self.scene.show_result(self.sudoku.graph)
        return True

    def reset_play_board(self):
        """
        Reset play board for new game
        """
        self.sudoku.reset_sudoku()
        self.scene.reset_play_board()
