"""
Author: Lukas Runt
Email: lrunt@students.zcu.cz
Date: November 21, 2023
Version: 1.0.0
Description: Communication module contains layout of the window and its texts
"""

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIntValidator, QFont
from PyQt5.QtWidgets import QWidget, QGridLayout, QLineEdit, QPushButton, QVBoxLayout, QTextEdit, \
    QHBoxLayout, QComboBox, QMessageBox, QLabel
from qtpy import QtGui

import BazeDat


class SudokuGUI(QWidget):
    """
    Sodoku graphical user interface represents layout of widgets
    """

    def __init__(self):
        super().__init__()
        self.grid_layout = QGridLayout()
        self.grid_layout.setSpacing(0)
        self.text_boxes = [None] * BazeDat.NUMBER_OF_TILES
        self.solve_button = QPushButton(solve_button_text, self)
        self.sudoku_layout = QVBoxLayout()
        self.explanation_layout = QVBoxLayout()
        self.list_widget = QComboBox(self)
        self.coordination_list = QComboBox(self)
        self.explanation_button = QPushButton(explanation_button_text, self)
        self.explanation_area = QTextEdit(self)
        self.layout = QHBoxLayout()
        self.initUI()

    def initUI(self):
        """
        Initialization of layout
        """
        self.create_sudoku_layout()
        self.create_explanation_layout()
        self.layout.addLayout(self.sudoku_layout)
        self.layout.addLayout(self.explanation_layout)

    def create_sudoku_layout(self):
        """
        Creates sudoku layout
        """
        # Coordinates
        for row in range(BazeDat.HEIGHT_OF_PLAY_BOARD):
            label = QLabel(str(row) + " ")
            label.setFixedWidth(10)
            self.grid_layout.addWidget(label, row + 1, 0)

        for column in range(len(BazeDat.alphabet)):
            label = QLabel(BazeDat.alphabet[column])
            label.setFixedHeight(20)
            label.setAlignment(Qt.AlignCenter)
            self.grid_layout.addWidget(label, 0, column + 1)

        # Play board
        for row in range(BazeDat.HEIGHT_OF_PLAY_BOARD):
            for col in range(BazeDat.WIDTH_OF_PLAY_BOARD):
                index = row * BazeDat.WIDTH_OF_PLAY_BOARD + col
                text_box = QLineEdit(self)
                text_box.setMaxLength(1)
                text_box.setAlignment(Qt.AlignCenter)

                int_validator = NoLeadingZeroValidator(1, 9)
                text_box.setValidator(int_validator)

                # Set fixed width and height to make it square
                text_box.setFixedWidth(60)
                text_box.setFixedHeight(60)

                font = text_box.font()
                font.setPointSize(16)  # Set the font size
                font.setFamily("Arial")  # Set the font family
                text_box.setFont(font)

                # Set borders for the 3x3 subgrids
                if row % 3 == 0:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-top: 2px solid black;")
                else:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-top: 1px solid gray;")
                if row % 3 == 2:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-bottom: 2px solid black;")
                else:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-bottom: 1px solid gray;")
                if col % 3 == 0:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-left: 2px solid black;")
                else:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-left: 1px solid gray;")
                if col % 3 == 2:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-right: 2px solid black;")
                else:
                    text_box.setStyleSheet(text_box.styleSheet() + "border-right: 1px solid gray;")

                # Set background color for every odd text field
                if index % 2 == 1:
                    text_box.setStyleSheet(text_box.styleSheet() + "background-color: lightblue;")

                self.text_boxes[index] = text_box
                self.grid_layout.addWidget(text_box, row + 1, col + 1)

        self.sudoku_layout.addLayout(self.grid_layout)
        self.sudoku_layout.addWidget(self.solve_button)

    def create_explanation_layout(self):
        """
        Creates explanation layout
        """
        self.list_widget.addItem(print_process)
        self.list_widget.addItem(print_explanation_of_field_solution)

        # Because I want to sort it A0, A1, A2....
        for col in range(BazeDat.HEIGHT_OF_PLAY_BOARD):
            for row in range(BazeDat.WIDTH_OF_PLAY_BOARD):
                index = row * BazeDat.WIDTH_OF_PLAY_BOARD + col
                self.coordination_list.addItem(get_field_coordinate(index))

        self.explanation_area.setReadOnly(True)
        self.explanation_button.setDisabled(True)
        self.coordination_list.setDisabled(True)
        self.list_widget.activated.connect(self.comboBoxActivated)

        lists_layout = QHBoxLayout()
        lists_layout.addWidget(self.list_widget)
        lists_layout.addWidget(self.coordination_list)

        self.explanation_layout.addLayout(lists_layout)
        self.explanation_layout.addWidget(self.explanation_button)
        self.explanation_layout.addWidget(self.explanation_area)

    def change_solve_button_text(self):
        """
        Changing button text to reset or solve
        """
        if self.solve_button.text() == solve_button_text:
            self.solve_button.setText(reset_button_text)
        else:
            self.solve_button.setText(solve_button_text)

    def set_text_box_bold(self, i):
        """
        Method setting bold style of text box
        :param i: index of text box in list
        """
        font = QFont()
        font.setBold(True)
        font.setPointSize(16)
        font.setFamily("Arial")
        self.text_boxes[i].setFont(font)

    def show_result(self, graph):
        """
        Method for showing right result
        :param graph: graph of sudoku with result
        """
        for i in range(len(self.text_boxes)):
            node = graph.nodes[i]
            if not node.defined:
                self.set_text_box_bold(i)
            self.text_boxes[i].setText(str(node.value))
            self.text_boxes[i].setReadOnly(True)
            self.explanation_button.setDisabled(False)

    def reset_play_board(self):
        """
        Method reset play board
        """
        self.explanation_button.setDisabled(True)
        self.explanation_area.setText("")
        for text_box in self.text_boxes:
            text_box.setText("")
            font = QFont()
            font.setBold(False)
            font.setPointSize(16)  # Set the font size
            font.setFamily("Arial")
            text_box.setFont(font)
            text_box.setReadOnly(False)

    def comboBoxActivated(self):
        """
        Changing disability of coordination list
        """
        if self.list_widget.currentText() == print_process:
            self.coordination_list.setDisabled(True)
        else:
            self.coordination_list.setDisabled(False)


class NoLeadingZeroValidator(QIntValidator):
    """
    Class to disable 0 input in the textbox
    """

    def validate(self, input_str, pos):
        result, input_str, pos = super().validate(input_str, pos)
        if input_str.startswith("0"):
            return QIntValidator.Invalid, input_str, pos
        else:
            return result, input_str, pos


def possible_only_one_number(step, coordinates, value):
    """
    Return text of one step, where is only one possible number to assign
    :param step: number of step
    :param coordinates: coordinates of the field
    :param value: value what has been assigned
    :return: created text
    """
    return f"Krok {step:2}: Do pole {get_field_coordinate(coordinates)} lze dolnit pouze číslo {value}. Doplníme tedy číslici {value}.\n"


def possible_more_numbers(step, coordinates, values, value):
    """
    Return text of one step, where is more than one possible numbers to assign
    :param step: number of step
    :param coordinates: coordinates of the field
    :param values: possible values to assign
    :param value: value what has been assigned
    :return: created text
    """
    return f"Krok {step:2}: Do pole {get_field_coordinate(coordinates)} lze dolnit jedno z následujících číslic: {values}. Doplníme první zmíněnou číslici {value}.\n"


def possible_more_numbers_some_tested(step, coordinates, values, tested_values, value):
    """
    Return text of one step, where is more than one possible numbers to assign and some of than has been tested
    :param step: number of step
    :param coordinates: coordinates of the field
    :param values: possible values to assign
    :param tested_values: values what algorithm try to test
    :param value: value what has been assigned
    :return: created text
    """
    return f"Krok {step:2}: Do pole {get_field_coordinate(coordinates)} lze dolnit jedno z následujících číslic: {values}. Algoritmus se pokusil projít číslice {tested_values} a zjistil, že zvolil špatně. Dolníme tedy číslici {value}.\n"


def value_entered_by_user(coordinates, value):
    """
    Text for explanation a value entered by user
    :param coordinates: coordinates of field
    :param value: value of the field
    :return: created text
    """
    return f"Hodnota pole {coordinates} byla definována uživatelem číslem {value}.\n"


def sudoku_before_solution(coordinates, step):
    """
    Text of solution before field
    :param step: step of the solution
    :param coordinates: coordinates of the field
    :return: created text
    """
    return f"Sudoku před řešením pole {coordinates} (Krok {step}):\n"


def show_error_message(message):
    """
    Method to show error message to the user
    :param message: Message that will be shown to the user
    """
    error_message = QMessageBox()
    error_message.setIcon(QMessageBox.Critical)
    error_message.setWindowIcon(QtGui.QIcon(error_message_box_icon))
    error_message.setText(message)
    error_message.setWindowTitle("Error")
    error_message.setStandardButtons(QMessageBox.Ok)
    error_message.exec_()


def get_field_coordinate(index):
    """
    Transform index of array to coordinate
    :param index: index in array
    :return: coordinates of index
    """
    row = index // BazeDat.WIDTH_OF_PLAY_BOARD
    column = BazeDat.alphabet[index % BazeDat.WIDTH_OF_PLAY_BOARD]
    return f"{column}{row}"


def get_explanation_for_no_color(where):
    """
    Text for no color connected with field
    :param where: where are no colors connected with field
    :return: created text
    """
    return f"{where} není žádné číslo.\n"


def get_explanation_for_one_color(where, color):
    """
    Text for one color connected with field
    :param where: where is only one color connected with field
    :param color: which color is connected with field
    :return: created text
    """
    return f"{where} je číslo {color}.\n"


def get_explanation_for_more_colors(where, colors):
    """
    Text for more than one color connected with field
    :param where: where is the colors connected with field
    :param colors: which colors are connected with field
    :return: created text
    """
    return f"{where} jsou čísla {colors}.\n"


def one_possible_color(color):
    """
    Text for explanation which color is possible in the field
    :param color: which color is possible
    :return: created text
    """
    return f"Z předchozích informací vyplívá, že do zvoleného pole může být dosazeno pouze číslo {color}.\n"


def more_possible_colors(colors):
    """
    Text for explanation which color are possible in the field
    :param colors: which color are possible
    :return: created text
    """
    return f"Z předchozích informací vyplívá, že do zvoleného pole mohou být dosazena čísla {colors}.\n"


def what_algorithm_assigned_simple(value):
    """
    Text for explanation which color algorithm has assigned
    :param value: value what algorithm has assigned
    :return: created text
    """
    return f"Algoritmus dolnil do sudoku číslici {value}.\n"


def what_algorithm_assigned_complicated(value, other_values):
    """
    Text for explanation which color algorithm has assigned
    :param value: value what algorithm has assigned
    :param other_values: other values what have algorithm tried
    :return: created text
    """
    return f"Algoritmus doplnil do sudoku číslici {value}, protože při dosazení číslic {other_values} nenašel řešení.\n"


# Titles of windows
main_window_title = "Řešitel sudoku"

# Texts for buttons
solve_button_text = "ŘEŠIT"
reset_button_text = "RESET"
explanation_button_text = "VYSVĚTLIT"

# Texts for explanation list
print_process = "Postup"
print_explanation_of_field_solution = "Vysvětli řešení pole"

# Texts for errors
error_no_solution = f"Toto sudoku nemá řešení."
error_bad_input = f"Toto sudoku nemá řešení, protože se v zadání objevuje porušení pravidel hry sudoku."

# Icons
main_window_icon = 'Data/sudoku_icon.png'
error_message_box_icon = 'Data/error_icon.png'

# Chars in sudoku
actual_position = "x"
blank_field = '\u2002'
