"""
Author: Lukas Runt
Email: lrunt@students.zcu.cz
Date: January 04, 2024
Version: 1.1.0
Description: Inference module, managing communication between modules
"""
import sys

from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget, QMenu, QFileDialog, QAction
from qtpy import QtGui

import BazeDat
import BazeZnalosti
import KomunikacniModul
import VysvetlovaciModul


def app():
    """
    Method creating main window of the application
    """
    application = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(application.exec_())


def get_integer_input(string, minimum, maximum):
    """
    Method controls correctness of user input
    Controls if input is integer, if yes then controls if integer is in interval
    :param string: Incoming string, what will be converted to integer
    :param minimum: minimal possible number
    :param maximum: maximal possible number
    :return: -1 = Wrong input, or number
    """
    try:
        integer_value = int(string)
        if minimum <= integer_value <= maximum:
            return integer_value
    except ValueError:
        return -1
    return -1


class MainWindow(QMainWindow):
    """
    Starting main window of application
    """

    def __init__(self):
        super().__init__()

        self.sudoku = BazeZnalosti.Sudoku()

        self.setWindowTitle(KomunikacniModul.main_window_title)
        self.setWindowIcon(QtGui.QIcon(KomunikacniModul.main_window_icon))
        self.setGeometry(100, 100, 1000, 500)

        self.central_widget = QWidget(self)
        self.scene = KomunikacniModul.SudokuGUI()

        self.setCentralWidget(self.central_widget)
        self.central_widget.setLayout(self.scene.layout)
        self.scene.solve_button.clicked.connect(self.solution_button_action)
        self.scene.explanation_button.clicked.connect(self.explanation_button_action)

        self.import_action = QAction(KomunikacniModul.form_device, self)
        self.load_default = QAction(KomunikacniModul.default_sudoku, self)
        self.create_menu_bar()

    def create_menu_bar(self):
        """
        Adding menu bar to main window
        """
        menu_bar = self.menuBar()
        import_menu = QMenu(KomunikacniModul.load, self)
        menu_bar.addMenu(import_menu)

        self.import_action.triggered.connect(self.show_dialog)
        import_menu.addAction(self.import_action)

        self.load_default.triggered.connect(self.load_default_sudoku)
        import_menu.addAction(self.load_default)

    def show_dialog(self):
        """
        Showing file browser window, showing only text files
        """
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog  # Optional, for better compatibility on macOS and Linux

        # Set the file filter to only allow .txt files
        file_filter = "Text files (*.txt)"
        path, _ = QFileDialog.getOpenFileName(self, 'Open file', '', file_filter, options=options)

        if path:
            self.parse_file(path)

    def parse_file(self, path):
        """
        Method tries to load data
        :param path: path to the file
        """
        i = 0
        with open(path, 'r') as file:
            for line in file:
                numbers = line.split(' ')
                if len(numbers) == BazeDat.WIDTH_OF_PLAY_BOARD:
                    for str_number in numbers:
                        number = get_integer_input(str_number, BazeDat.INPUT_MINIMUM, BazeDat.INPUT_MAXIMUM)
                        if number == -1:
                            KomunikacniModul.show_error_message(KomunikacniModul.invalid_file_format)
                            return
                        else:
                            BazeDat.sudoku[i] = number
                            i += 1
        if i != BazeDat.NUMBER_OF_TILES:
            KomunikacniModul.show_error_message(KomunikacniModul.invalid_file_format)
        else:
            self.scene.load_result(BazeDat.sudoku)

    def load_default_sudoku(self):
        """
        Loading default sudoku to the play board
        """
        self.scene.load_result(BazeDat.default_sudoku)

    def solution_button_action(self):
        """
        Choose which action will be performed
        """
        if self.scene.solve_button.text() == KomunikacniModul.solve_button_text:
            if self.solve_sudoku():
                self.scene.change_solve_button_text()
                self.load_default.setEnabled(False)
                self.import_action.setEnabled(False)
        else:
            self.reset_play_board()
            self.load_default.setEnabled(True)
            self.import_action.setEnabled(True)
            self.scene.change_solve_button_text()

    def explanation_button_action(self):
        """
        Handling the explanation button action
        """
        if self.scene.list_widget.currentText() == KomunikacniModul.print_process:
            explanation_text = VysvetlovaciModul.explain_solution_procedure(self.sudoku.stack)
            self.scene.explanation_area.setText(explanation_text)
        else:
            coordination = self.scene.coordination_list.currentText()
            explanation_text = VysvetlovaciModul.explain_tile_value_choice(coordination, self.sudoku)
            self.scene.explanation_area.setText(explanation_text)

    def solve_sudoku(self):
        """
        Sending data from graphical interface to database
        :return: True - Sudoku solved, False - Solving sudoku failed
        """
        for i in range(len(self.scene.text_boxes)):
            value = self.scene.text_boxes[i].text()
            if value == "" or value == " ":
                BazeDat.sudoku[i] = 0
            else:
                BazeDat.sudoku[i] = int(value)
        self.sudoku.fill_graph_with_given_information()
        if not self.sudoku.control_input_data():
            KomunikacniModul.show_error_message(KomunikacniModul.error_bad_input)
            return False
        if not self.sudoku.find_solution_of_sudoku():
            KomunikacniModul.show_error_message(KomunikacniModul.error_no_solution)
            return False
        self.scene.show_result(self.sudoku.graph)
        return True

    def reset_play_board(self):
        """
        Reset play board for new game
        """
        self.sudoku.reset_sudoku()
        self.scene.reset_play_board()
